from __future__ import annotations

import spacy.symbols as POS
from rapidfuzz.distance import Levenshtein
from spacy.tokens import Token

from vnerrant.components.classifier import BaseClassifer
from vnerrant.components.en.constants import ChildrenErrorType
from vnerrant.components.en.constants import dependency_resources
from vnerrant.components.en.constants import language_resources
from vnerrant.components.en.constants import ParentErrorType
from vnerrant.components.en.constants import pos_resources
from vnerrant.constants import Operator
from vnerrant.constants import SeparatorTypes
from vnerrant.model.edit import Edit


class Classifier(BaseClassifer):
    """
    Classifier class for English
    """

    def __init__(self):
        super().__init__()

    @staticmethod
    def only_orth_change(o_toks: list[Token], c_toks: list[Token]) -> bool:
        """
        Check if the difference between two lists of tokens is only whitespace or case.
        :param o_toks: A list of spacy tokens.
        :param c_toks: A list of spacy tokens.
        :return: Boolean.
        """
        o_join = ''.join([o.lower_ for o in o_toks])
        c_join = ''.join([c.lower_ for c in c_toks])
        if o_join == c_join:
            return True
        return False

    @staticmethod
    def exact_reordering(o_toks: list[Token], c_toks: list[Token]) -> bool:
        """
        Check if the difference between two lists of tokens is only reordering.
        :param o_toks: A list of spacy tokens.
        :param c_toks: A list of spacy tokens.
        :return: Boolean.
        """
        # Sorting lets us keep duplicates.
        o_set = sorted([o.lower_ for o in o_toks])
        c_set = sorted([c.lower_ for c in c_toks])
        if o_set == c_set:
            return True
        return False

    @staticmethod
    def preceded_by_aux(o_toks: list[Token], c_toks: list[Token]) -> bool:
        """
        Check if two tokens are preceded by an auxiliary verb.
        :param o_toks: A list of spacy tokens.
        :param c_toks: A list of spacy tokens.
        :return: Boolean.
        """
        # If the toks are aux, we need to check if they are the first aux.
        if o_toks[0].dep_.startswith('aux') and c_toks[0].dep_.startswith('aux'):
            # Find the parent verb
            o_head = o_toks[0].head
            c_head = c_toks[0].head
            # Find the children of the parent
            o_children = o_head.children
            c_children = c_head.children
            # Check the orig children.
            for o_child in o_children:
                # Look at the first aux...
                if o_child.dep_.startswith('aux'):
                    # Check if the string matches o_tok
                    if o_child.text != o_toks[0].text:
                        # If it doesn't, o_tok is not first so check cor
                        for c_child in c_children:
                            # Find the first aux in cor...
                            if c_child.dep_.startswith('aux'):
                                # If that doesn't match either, neither are first aux
                                if c_child.text != c_toks[0].text:
                                    return True
                                # Break after the first cor aux
                                break
                    # Break after the first orig aux.
                    break
        # Otherwise, the toks are main verbs so we need to look for any aux.
        else:
            o_deps = [o_dep.dep_ for o_dep in o_toks[0].children]
            c_deps = [c_dep.dep_ for c_dep in c_toks[0].children]
            if 'aux' in o_deps or 'auxpass' in o_deps:
                if 'aux' in c_deps or 'auxpass' in c_deps:
                    return True
        return False

    @staticmethod
    def get_pos_and_dep(toks: list[Token]) -> tuple[list[str], list[str]]:
        """
        Extract POS and dependency parse info from a list of spacy tokens.
        :param toks: A list of spacy tokens.
        :return: A tuple of lists of POS and dependency parse tags.
        """
        pos = []
        dep = []
        for tok in toks:
            pos.append(pos_resources.pos_map[tok.tag_])
            dep.append(tok.dep_)
        return pos, dep

    def is_only_orth_change(self, o_toks: list[Token], c_toks: list[Token]) -> bool:
        """
        Check if the difference between two lists of tokens is only whitespace or case.
        :param o_toks: A list of spacy tokens.
        :param c_toks: A list of spacy tokens.
        :return: Boolean.
        """
        return self.only_orth_change(o_toks, c_toks)

    def is_exact_reordering(self, o_toks: list[Token], c_toks: list[Token]) -> bool:
        """
        Check if the difference between two lists of tokens is only reordering.
        :param o_toks: A list of spacy tokens.
        :param c_toks: A list of spacy tokens.
        :return: Boolean.
        """
        return self.exact_reordering(o_toks, c_toks)

    def is_preceded_by_aux(self, o_toks: list[Token], c_toks: list[Token]) -> bool:
        """
        Check if two tokens are preceded by an auxiliary verb.
        :param o_toks: A list of spacy tokens.
        :param c_toks: A list of spacy tokens.
        :return: Boolean.
        """
        return self.preceded_by_aux(o_toks, c_toks)

    def _get_one_sided_type(self, toks: list[Token]) -> str:
        """
        Classify an edit with only one side of tokens.
        :param toks: A list of spacy tokens.
        :return: An error type string.
        """
        # Special cases
        if len(toks) == 1:
            # Possessive noun suffixes; e.g. ' -> 's
            if toks[0].tag_ == 'POS':
                return (
                    ParentErrorType.NOUN
                    + SeparatorTypes.COLON
                    + ChildrenErrorType.POSSESSIVE
                )
            # Contractions. Rule must come after possessive
            if toks[0].lower_ in dependency_resources.conts:
                return ParentErrorType.CONTRACTION
            # Infinitival "to" is treated as part of a verb form
            if (
                toks[0].lower_ == 'to'
                and toks[0].pos == POS.PART
                and toks[0].dep_ != 'prep'
            ):
                return (
                    ParentErrorType.VERB + SeparatorTypes.COLON + ChildrenErrorType.FORM
                )
        # Extract pos tags and parse info from the toks
        pos_list, dep_list = self.get_pos_and_dep(toks)
        # Auxiliary verbs
        if set(dep_list).issubset({'aux', 'auxpass'}):
            return ParentErrorType.VERB + SeparatorTypes.COLON + ChildrenErrorType.TENSE
        # POS-based tags. Ignores rare, uninformative categories
        if len(set(pos_list)) == 1 and pos_list[0] not in pos_resources.rare_pos:
            return pos_list[0]  # TODO: check if this is correct
        # More POS-based tags using special dependency labels
        if (
            len(set(dep_list)) == 1
            and dep_list[0] in dependency_resources.dep_map.keys()
        ):
            return dependency_resources.dep_map[dep_list[0]]
        # To-infinitives and phrasal verbs
        if set(pos_list) == {'PART', 'VERB'}:
            return ParentErrorType.VERB
        # Tricky cases
        else: # TODO: Get more specific error type
            return ParentErrorType.OTHER

    def _get_two_sided_type(self, o_toks: list[Token], c_toks: list[Token]) -> str:
        """
        Classify an edit with two sides of tokens.
        :param o_toks: A list of spacy tokens.
        :param c_toks: A list of spacy tokens.
        :return: An error type string.
        """
        # Extract pos tags and parse info from the toks as lists
        o_pos, o_dep = self.get_pos_and_dep(o_toks)
        c_pos, c_dep = self.get_pos_and_dep(c_toks)

        # Orthography; i.e. whitespace and/or case errors.
        if self.is_only_orth_change(o_toks, c_toks):
            return ParentErrorType.ORTHOGRAPHY
        # Word Order; only matches exact reordering.
        if self.is_exact_reordering(o_toks, c_toks):
            return ParentErrorType.WORD_ORDER

        # 1:1 replacements (very common)
        if len(o_toks) == len(c_toks) == 1:
            # 1. SPECIAL CASES
            # Possessive noun suffixes; e.g. ' -> 's
            if o_toks[0].tag_ == 'POS' or c_toks[0].tag_ == 'POS':
                return (
                    ParentErrorType.NOUN
                    + SeparatorTypes.COLON
                    + ChildrenErrorType.POSSESSIVE
                )
            # Contraction. Rule must come after possessive.
            if (
                o_toks[0].lower_ in dependency_resources.conts
                or c_toks[0].lower_ in dependency_resources.conts
            ) and o_pos == c_pos:
                return ParentErrorType.CONTRACTION
            # Special auxiliaries in contractions (1); e.g. ca -> can, wo -> will
            # Rule was broken in V1. Turned off this fix for compatibility.
            if (
                o_toks[0].lower_ in dependency_resources.aux_conts
                and c_toks[0].lower_ == dependency_resources.aux_conts[o_toks[0].lower_]
            ) or (
                c_toks[0].lower_ in dependency_resources.aux_conts
                and o_toks[0].lower_ == dependency_resources.aux_conts[c_toks[0].lower_]
            ):
                return ParentErrorType.CONTRACTION
            # Special auxiliaries in contractions (2); e.g. ca -> could, wo -> should
            if (
                o_toks[0].lower_ in dependency_resources.aux_conts
                or c_toks[0].lower_ in dependency_resources.aux_conts
            ):
                return (
                    ParentErrorType.VERB
                    + SeparatorTypes.COLON
                    + ChildrenErrorType.TENSE
                )
            # Special: "was" and "were" are the only past tense SVA
            if {o_toks[0].lower_, c_toks[0].lower_} == {'was', 'were'}:
                return (
                    ParentErrorType.VERB
                    + SeparatorTypes.COLON
                    + ChildrenErrorType.SUBJECT_VERB_AGREEMENT
                )

            # 2. SPELLING AND INFLECTION
            # Only check alphabetical strings on the original side
            # Spelling errors take precedence over POS errors; this rule is ordered
            if o_toks[0].text.isalpha():
                # Check a GB English dict for both orig and lower case.
                # E.g. "cat" is in the dict, but "Cat" is not.
                if (
                    o_toks[0].text not in language_resources.spell
                    and o_toks[0].lower_ not in language_resources.spell
                ):
                    # Check if both sides have a common lemma
                    if o_toks[0].lemma == c_toks[0].lemma:
                        # Inflection; often count vs mass nouns or e.g. got vs getted
                        if o_pos == c_pos and o_pos[0] in {'NOUN', 'VERB'}:
                            return (
                                o_pos[0]
                                + SeparatorTypes.COLON
                                + ChildrenErrorType.INFLECTION
                            )
                        # Unknown morphology; i.e. we cannot be more specific.
                        else:
                            return ParentErrorType.MORPHOLOGY
                    # Use string similarity to detect true spelling errors.
                    else:
                        # Normalised Lev distance works better than Lev ratio
                        str_sim = Levenshtein.normalized_similarity(
                            o_toks[0].lower_,
                            c_toks[0].lower_,
                        )
                        # WARNING: THIS IS AN APPROXIMATION.
                        # Thresholds tuned manually on FCE_train + W&I_train
                        # str_sim > 0.55 is almost always a true spelling error
                        if str_sim > 0.55:
                            return ParentErrorType.SPELLING
                        # Special scores for shorter sequences are usually SPELL
                        if str_sim == 0.5 or round(str_sim, 3) == 0.333:
                            # Short strings are more likely to be spell: eles -> else
                            if len(o_toks[0].text) <= 4 and len(c_toks[0].text) <= 4:
                                return ParentErrorType.SPELLING
                        # The remainder are usually word choice: amounght -> number
                        # Classifying based on cor_pos alone is generally enough.
                        if c_pos[0] not in pos_resources.rare_pos:
                            return c_pos[0]
                        # Anything that remains is OTHER
                        else:
                            return ParentErrorType.OTHER

            # 3. MORPHOLOGY
            # Only ADJ, ADV, NOUN and VERB can have inflectional changes.
            if (
                o_toks[0].lemma == c_toks[0].lemma
                and o_pos[0] in pos_resources.open_pos2
                and c_pos[0] in pos_resources.open_pos2
            ):
                # Same POS on both sides
                if o_pos == c_pos:
                    # Adjective form; e.g. comparatives
                    if o_pos[0] == 'ADJ':
                        return (
                            ParentErrorType.ADJECTIVE
                            + SeparatorTypes.COLON
                            + ChildrenErrorType.FORM
                        )
                    # Noun number
                    if o_pos[0] == 'NOUN':
                        return (
                            ParentErrorType.NOUN
                            + SeparatorTypes.COLON
                            + ChildrenErrorType.NUMBER
                        )
                    # Verbs - various types
                    if o_pos[0] == 'VERB':
                        # NOTE: These rules are carefully ordered.
                        # Use the dep parse to find some form errors.
                        # Main verbs preceded by aux cannot be tense or SVA.
                        if self.is_preceded_by_aux(o_toks, c_toks):
                            return (
                                ParentErrorType.VERB
                                + SeparatorTypes.COLON
                                + ChildrenErrorType.FORM
                            )
                        # Use fine PTB tags to find various errors.
                        # FORM errors normally involve VBG or VBN.
                        if o_toks[0].tag_ in {'VBG', 'VBN'} or c_toks[0].tag_ in {
                            'VBG',
                            'VBN',
                        }:
                            return (
                                ParentErrorType.VERB
                                + SeparatorTypes.COLON
                                + ChildrenErrorType.FORM
                            )
                        # Of what's left, TENSE errors normally involved VBD.
                        if o_toks[0].tag_ == 'VBD' or c_toks[0].tag_ == 'VBD':
                            return (
                                ParentErrorType.VERB
                                + SeparatorTypes.COLON
                                + ChildrenErrorType.TENSE
                            )
                        # Of what's left, SVA errors normally involve VBZ.
                        if o_toks[0].tag_ == 'VBZ' or c_toks[0].tag_ == 'VBZ':
                            return (
                                ParentErrorType.VERB
                                + SeparatorTypes.COLON
                                + ChildrenErrorType.SUBJECT_VERB_AGREEMENT
                            )
                        # Any remaining aux verbs are called TENSE.
                        if o_dep[0].startswith('aux') and c_dep[0].startswith('aux'):
                            return (
                                ParentErrorType.VERB
                                + SeparatorTypes.COLON
                                + ChildrenErrorType.TENSE
                            )
                # Use dep labels to find some more ADJ:FORM
                if set(o_dep + c_dep).issubset({'acomp', 'amod'}):
                    return (
                        ParentErrorType.ADJECTIVE
                        + SeparatorTypes.COLON
                        + ChildrenErrorType.FORM
                    )
                # Adj to plural noun is usually noun number; e.g. musical -> musicals.
                if o_pos[0] == 'ADJ' and c_toks[0].tag_ == 'NNS':
                    return (
                        ParentErrorType.NOUN
                        + SeparatorTypes.COLON
                        + ChildrenErrorType.NUMBER
                    )
                # For remaining verb errors (rare), rely on c_pos
                if c_toks[0].tag_ in {'VBG', 'VBN'}:
                    return (
                        ParentErrorType.VERB
                        + SeparatorTypes.COLON
                        + ChildrenErrorType.FORM
                    )
                if c_toks[0].tag_ == 'VBD':
                    return (
                        ParentErrorType.VERB
                        + SeparatorTypes.COLON
                        + ChildrenErrorType.TENSE
                    )
                if c_toks[0].tag_ == 'VBZ':
                    return (
                        ParentErrorType.VERB
                        + SeparatorTypes.COLON
                        + ChildrenErrorType.SUBJECT_VERB_AGREEMENT
                    )
                # Tricky cases that all have the same lemma.
                else:
                    return ParentErrorType.MORPHOLOGY
            # Derivational morphology.
            if (
                language_resources.stemmer.stem(o_toks[0].text)
                == language_resources.stemmer.stem(c_toks[0].text)
                and o_pos[0] in pos_resources.open_pos2
                and c_pos[0] in pos_resources.open_pos2
            ):
                return ParentErrorType.MORPHOLOGY

            # 4. GENERAL
            # Auxiliaries with different lemmas
            if o_dep[0].startswith('aux') and c_dep[0].startswith('aux'):
                return (
                    ParentErrorType.VERB
                    + SeparatorTypes.COLON
                    + ChildrenErrorType.TENSE
                )
            # POS-based tags. Some of these are context sensitive mispellings.
            if o_pos == c_pos and o_pos[0] not in pos_resources.rare_pos:
                return o_pos[0]
            # Some dep labels map to POS-based tags.
            if o_dep == c_dep and o_dep[0] in dependency_resources.dep_map.keys():
                return dependency_resources.dep_map[o_dep[0]]
            # Phrasal verb particles.
            if set(o_pos + c_pos) == {'PART', 'PREP'} or set(o_dep + c_dep) == {
                'prt',
                'prep',
            }:
                return ParentErrorType.PARTICLE
            # Can use dep labels to resolve DET + PRON combinations.
            if set(o_pos + c_pos) == {'DET', 'PRON'}:
                # DET cannot be a subject or object.
                if c_dep[0] in {'nsubj', 'nsubjpass', 'dobj', 'pobj'}:
                    return ParentErrorType.PRONOUN
                # "poss" indicates possessive determiner
                if c_dep[0] == 'poss':
                    return ParentErrorType.DETERMINER
            # NUM and DET are usually DET; e.g. a <-> one
            if set(o_pos + c_pos) == {'NUM', 'DET'}:
                return ParentErrorType.DETERMINER
            # Special: other <-> another
            if {o_toks[0].lower_, c_toks[0].lower_} == {'other', 'another'}:
                return ParentErrorType.DETERMINER
            # Special: your (sincerely) -> yours (sincerely)
            if o_toks[0].lower_ == 'your' and c_toks[0].lower_ == 'yours':
                return ParentErrorType.PRONOUN
            # Special: no <-> not; this is very context sensitive
            if {o_toks[0].lower_, c_toks[0].lower_} == {'no', 'not'}:
                return ParentErrorType.OTHER

            # 5. STRING SIMILARITY
            # These rules are quite language specific.
            if o_toks[0].text.isalpha() and c_toks[0].text.isalpha():
                # Normalised Lev distance works better than Lev ratio
                str_sim = Levenshtein.normalized_similarity(
                    o_toks[0].lower_,
                    c_toks[0].lower_,
                )
                # WARNING: THIS IS AN APPROXIMATION.
                # Thresholds tuned manually on FCE_train + W&I_train
                # A. Short sequences are likely to be SPELL or function word errors
                if len(o_toks[0].text) == 1:
                    # i -> in, a -> at
                    if len(c_toks[0].text) == 2 and str_sim == 0.5:
                        return ParentErrorType.SPELLING
                if len(o_toks[0].text) == 2:
                    # in -> is, he -> the, to -> too
                    if 2 <= len(c_toks[0].text) <= 3 and str_sim >= 0.5:
                        return ParentErrorType.SPELLING
                if len(o_toks[0].text) == 3:
                    # Special: the -> that (relative pronoun)
                    if o_toks[0].lower_ == 'the' and c_toks[0].lower_ == 'that':
                        return ParentErrorType.PRONOUN
                    # Special: all -> everything
                    if o_toks[0].lower_ == 'all' and c_toks[0].lower_ == 'everything':
                        return ParentErrorType.PRONOUN
                    # off -> of, too -> to, out -> our, now -> know
                    if 2 <= len(c_toks[0].text) <= 4 and str_sim >= 0.5:
                        return ParentErrorType.SPELLING
                # B. Longer sequences are also likely to include content word errors
                if len(o_toks[0].text) == 4:
                    # Special: that <-> what
                    if {o_toks[0].lower_, c_toks[0].lower_} == {'that', 'what'}:
                        return ParentErrorType.PRONOUN
                    # Special: well <-> good
                    if {o_toks[0].lower_, c_toks[0].lower_} == {
                        'good',
                        'well',
                    } and c_pos[0] not in pos_resources.rare_pos:
                        return c_pos[0]
                    # knew -> new,
                    if len(c_toks[0].text) == 3 and str_sim > 0.5:
                        return ParentErrorType.SPELLING
                    # then <-> than, form -> from
                    if len(c_toks[0].text) == 4 and str_sim >= 0.5:
                        return ParentErrorType.SPELLING
                    # gong -> going, hole -> whole
                    if len(c_toks[0].text) == 5 and str_sim == 0.8:
                        return ParentErrorType.SPELLING
                    # high -> height, west -> western
                    if (
                        len(c_toks[0].text) > 5
                        and str_sim > 0.5
                        and c_pos[0] not in pos_resources.rare_pos
                    ):
                        return c_pos[0]
                if len(o_toks[0].text) == 5:
                    # Special: after -> later
                    if {o_toks[0].lower_, c_toks[0].lower_} == {
                        'after',
                        'later',
                    } and c_pos[0] not in pos_resources.rare_pos:
                        return c_pos[0]
                    # where -> were, found -> fund
                    if len(c_toks[0].text) == 4 and str_sim == 0.8:
                        return ParentErrorType.SPELLING
                    # thing <-> think, quite -> quiet, their <-> there
                    if len(c_toks[0].text) == 5 and str_sim >= 0.6:
                        return ParentErrorType.SPELLING
                    # house -> domestic, human -> people
                    if (
                        len(c_toks[0].text) > 5
                        and c_pos[0] not in pos_resources.rare_pos
                    ):
                        return c_pos[0]
                # C. Longest sequences include MORPH errors
                if len(o_toks[0].text) > 5 and len(c_toks[0].text) > 5:
                    # Special: therefor -> therefore
                    if (
                        o_toks[0].lower_ == 'therefor'
                        and c_toks[0].lower_ == 'therefore'
                    ):
                        return ParentErrorType.SPELLING
                    # Special: though <-> thought
                    if {o_toks[0].lower_, c_toks[0].lower_} == {'though', 'thought'}:
                        return ParentErrorType.SPELLING
                    # Morphology errors: stress -> stressed, health -> healthy
                    if (
                        o_toks[0].text.startswith(c_toks[0].text)
                        or c_toks[0].text.startswith(o_toks[0].text)
                    ) and str_sim >= 0.66:
                        return ParentErrorType.MORPHOLOGY
                    # Spelling errors: exiting -> exciting, wether -> whether
                    if str_sim > 0.8:
                        return ParentErrorType.SPELLING
                    # Content word errors: learning -> studying, transport -> travel
                    if str_sim < 0.55 and c_pos[0] not in pos_resources.rare_pos:
                        return c_pos[0]
                    # NOTE: Errors between 0.55 and 0.8 are a mix of SPELL, MORPH and POS
            # Tricky cases
            else:
                return ParentErrorType.OTHER

        # Multi-token replacements (uncommon)
        # All auxiliaries
        if set(o_dep + c_dep).issubset({'aux', 'auxpass'}):
            return ParentErrorType.VERB + SeparatorTypes.COLON + ChildrenErrorType.TENSE
        # All same POS
        if len(set(o_pos + c_pos)) == 1:
            # Final verbs with the same lemma are tense; e.g. eat -> has eaten
            if o_pos[0] == 'VERB' and o_toks[-1].lemma == c_toks[-1].lemma:
                return (
                    ParentErrorType.VERB
                    + SeparatorTypes.COLON
                    + ChildrenErrorType.TENSE
                )
            # POS-based tags.
            elif o_pos[0] not in pos_resources.rare_pos:
                return o_pos[0]
        # All same special dep labels.
        if (
            len(set(o_dep + c_dep)) == 1
            and o_dep[0] in dependency_resources.dep_map.keys()
        ):
            return dependency_resources.dep_map[o_dep[0]]
        # Infinitives, gerunds, phrasal verbs.
        if set(o_pos + c_pos) == {'PART', 'VERB'}:
            # Final verbs with the same lemma are form; e.g. to eat -> eating
            if o_toks[-1].lemma == c_toks[-1].lemma:
                return (
                    ParentErrorType.VERB + SeparatorTypes.COLON + ChildrenErrorType.FORM
                )
            # Remaining edits are often verb; e.g. to eat -> consuming, look at -> see
            else:
                return ParentErrorType.VERB
        # Possessive nouns; e.g. friends -> friend 's
        if (o_pos == ['NOUN', 'PART'] or c_pos == ['NOUN', 'PART']) and o_toks[
            0
        ].lemma == c_toks[0].lemma:
            return (
                ParentErrorType.NOUN
                + SeparatorTypes.COLON
                + ChildrenErrorType.POSSESSIVE
            )
        # Adjective forms with "most" and "more"; e.g. more free -> freer
        if (
            (
                o_toks[0].lower_ in {'most', 'more'}
                or c_toks[0].lower_ in {'most', 'more'}
            )
            and o_toks[-1].lemma == c_toks[-1].lemma
            and len(o_toks) <= 2
            and len(c_toks) <= 2
        ):
            return (
                ParentErrorType.ADJECTIVE
                + SeparatorTypes.COLON
                + ChildrenErrorType.FORM
            )

        # Tricky cases.
        else:
            return ParentErrorType.OTHER

    def classify(self, edit: Edit) -> Edit:
        """
        Classify an edit into a specific error type with preprocessing the delimiters in edit.
        :param edit: An Edit object.
        :return: The error type of the edit.
        """
        clone_edit = edit.copy()

        if clone_edit.original.text.strip() or clone_edit.corrected.text.strip():
            clone_edit.original.strip()
            clone_edit.corrected.strip()

        clone_edit = self._classify(clone_edit)

        edit.edit_type = clone_edit.edit_type
        return edit

    def _classify(self, edit: Edit) -> Edit:
        """
        Classify an edit into a specific error type.
        :param edit: An Edit object.
        :return: The error type of the edit.
        """
        # Nothing to nothing is a detected but not corrected edit
        if not edit.original.tokens and not edit.corrected.tokens:
            edit.edit_type = ParentErrorType.UNK
        # Missing
        elif not edit.original.tokens and edit.corrected.tokens:
            op = Operator.MISSING
            cat = self._get_one_sided_type(edit.corrected.tokens)
            edit.edit_type = op + SeparatorTypes.COLON + cat
        # Unnecessary
        elif edit.original.tokens and not edit.corrected.tokens:
            op = Operator.UNNECESSARY
            cat = self._get_one_sided_type(edit.original.tokens)
            edit.edit_type = op + SeparatorTypes.COLON + cat
            if edit.corrected.text != '':
                op = Operator.REPLACE
                edit.edit_type = op + SeparatorTypes.COLON + cat
        # Replacement and special cases
        else:
            # Same to same is a detected but not corrected edit
            if edit.original.text == edit.corrected.text:
                # ' hello' -> 'hello ' # TODO: Dicussion needed for this case
                # edit.edit_type = ParentErrorType.UNK
                edit.edit_type = Operator.REPLACE + SeparatorTypes.COLON + ParentErrorType.OTHER
            # Special: Ignore case change at the end of multi token edits
            # E.g. [Doctor -> The doctor], [, since -> . Since]
            # Classify the edit as if the last token wasn't there
            elif edit.original.tokens[-1].lower == edit.corrected.tokens[-1].lower and (
                len(edit.original.tokens) > 1 or len(edit.corrected.tokens) > 1
            ):
                # Store a copy of the full orig and cor toks
                all_o_toks = edit.original.tokens[:]
                all_c_toks = edit.corrected.tokens[:]
                # Truncate the instance toks for classification
                edit.original.tokens = edit.original.tokens[:-1]
                edit.corrected.tokens = edit.corrected.tokens[:-1]
                # Classify the truncated edit
                edit = self._classify(edit)
                # Restore the full orig and cor toks
                edit.original.tokens = all_o_toks
                edit.corrected.tokens = all_c_toks
            # Replacement
            else:
                op = Operator.REPLACE
                cat = self._get_two_sided_type(
                    edit.original.tokens, edit.corrected.tokens)
                # if cat == ParentErrorType.CONTRACTION and "'" in edit.original.text:
                #     edit.corrected.text = " " + edit.corrected.text
                edit.edit_type = op + SeparatorTypes.COLON + cat
        return edit
